#include <QTime>
#include <QDebug>
#include "Map.h"

CMap  map;

//---------------------------------------------------------------------------------------------
CWall::CWall()
{
    mType = 0;
}

//---------------------------------------------------------------------------------------------
CWall::~CWall()
{
    deleteParams();
}

//---------------------------------------------------------------------------------------------
void CWall::deleteParams()
{
    for (size_t i = 0; i < mParams.size(); ++i)
        delete mParams[i];
    mParams.clear();
}

//---------------------------------------------------------------------------------------------
void CWall::setType(uint8_t type)
{
    if (mType == type)
        return;

    mType = type;

    // allocate the params
    deleteParams();
    std::vector<CParamType>& paramTypes = map.mWallsParams[type];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        CParam* newParam = NULL;

        switch (paramTypes[i].mType)
        {
            case eParamOrnate:
                newParam = (CParam*)(new CParamOrnate);
                break;

            default:
                break;
        }
        mParams.push_back(newParam);
    }
}

//---------------------------------------------------------------------------------------------
uint8_t CWall::getType()
{
    return  mType;
}

//---------------------------------------------------------------------------------------------
// assignment operator used for clipboard functions and copy/cut/paste tools
CWall&  CWall::operator=(CWall& rhs)
{
    // copy the type and alloc the params
    setType(rhs.mType);

    // copy the params
    for (size_t i = 0; i < mParams.size(); ++i)
    {
        switch (mParams[i]->mType)
        {
            case eParamOrnate:
            {
                CParamOrnate*   lParam = (CParamOrnate*)mParams[i];
                CParamOrnate*   rParam = (CParamOrnate*)rhs.mParams[i];
                lParam->mValue = rParam->mValue;
            }
            break;

            default:
                break;
        }
    }
    return *this;
}

//---------------------------------------------------------------------------------------------
void  CWall::load(FILE* handle)
{
    uint8_t type;
    fread(&type, 1, 1, handle);
    setType(type);

    std::vector<CParamType>& paramTypes = map.mWallsParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            case eParamOrnate:
                {
                    uint8_t value;
                    fread(&value, 1, 1, handle);
                    ((CParamOrnate*)mParams[i])->mValue = value;
                }
                break;

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
void  CWall::save(FILE* handle)
{
    fwrite(&mType, 1, 1, handle);

    std::vector<CParamType>& paramTypes = map.mWallsParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            case eParamOrnate:
                {
                    uint8_t value;

                    if (i < mParams.size())
                        value = ((CParamOrnate*)mParams[i])->mValue;
                    else
                        value = 0;

                    fwrite(&value, 1, 1, handle);
                }
                break;

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
CTile::CTile()
{
    mType = 0;
}

//---------------------------------------------------------------------------------------------
// assignment operator used for clipboard functions and copy/cut/paste tools
CTile&  CTile::operator=(CTile& rhs)
{
    mType = rhs.mType;

    for (int i = 0; i < eWallSideMax; ++i)
        mWalls[i] = rhs.mWalls[i];

    return *this;
}

//---------------------------------------------------------------------------------------------
CMap::CMap()
{
    // default values
    mSize = CVec2(15, 10);
    mTiles = NULL;
}

//---------------------------------------------------------------------------------------------
CMap::~CMap()
{
    if (mTiles != NULL)
        delete[] mTiles;

    mWallsParams.clear();
}

//---------------------------------------------------------------------------------------------
void CMap::resize()
{
    if (mTiles != NULL)
        delete[] mTiles;

    mTiles = new CTile[mSize.x * mSize.y];
}

//---------------------------------------------------------------------------------------------
void   CMap::setTile(CVec2 pos, uint8_t type)
{
    if (pos.x >= 0 && pos.x < mSize.x &&
        pos.y >= 0 && pos.y < mSize.y)
    {
        getTile(pos)->mType = type;
    }
}

//---------------------------------------------------------------------------------------------
void   CMap::setWall(CVec2 pos, EWallSide side, uint8_t type)
{
    if (pos.x >= 0 && pos.x < mSize.x &&
        pos.y >= 0 && pos.y < mSize.y)
    {
        getTile(pos)->mWalls[side].setType(type);
    }
}

//---------------------------------------------------------------------------------------------
CTile*  CMap::getTile(CVec2 pos)
{
    if (pos.x < 0 || pos.x >= mSize.x ||
        pos.y < 0 || pos.y >= mSize.y)
        return NULL;

    return(&mTiles[pos.y * mSize.x + pos.x]);
}

//---------------------------------------------------------------------------------------------
bool CMap::isLoaded()
{
    return mTiles != NULL;
}

//---------------------------------------------------------------------------------------------
char    gFileHeader[] = "DMMP"; // Dungeon Master MaP
#define CURRENT_VERSION     1
#define CURRENT_REVISION    1

void CMap::save(char* fileName)
{
    FILE*       handle = fopen(fileName, "wb");
    uint16_t    ver = CURRENT_VERSION;
    uint16_t    rev = CURRENT_REVISION;

    if (handle != NULL)
    {
        // header, version, revision
        fwrite(gFileHeader, 1, 4, handle);
        fwrite(&ver, 2, 1, handle);
        fwrite(&rev, 2, 1, handle);

        // map size
        fwrite(&mSize.x, 4, 1, handle);
        fwrite(&mSize.y, 4, 1, handle);

        // map data
        for (int y = 0; y < mSize.y; ++y)
            for (int x = 0; x < mSize.x; ++x)
            {
                CTile*  t = getTile(CVec2(x, y));
                fwrite(&t->mType, 1, 1, handle);

                for (int side = 0; side < eWallSideMax; ++side)
                    t->mWalls[side].save(handle);
            }

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
void CMap::load(char* fileName)
{
    FILE*       handle = fopen(fileName, "rb");
    char        header[4 + 1];
    uint16_t    ver;
    uint16_t    rev;

    if (handle != NULL)
    {
        // header, version, revision
        fread(header, 1, 4, handle);
        header[4] = 0;
        fread(&ver, 2, 1, handle);
        fread(&rev, 2, 1, handle);

#ifndef CONVERT_MAP
        // check if it's the right version
        if (strcmp(header, gFileHeader) != 0 ||
            ver != CURRENT_VERSION ||
            rev != CURRENT_REVISION)
        {
            fclose(handle);
            return;
        }
#endif

        // map size
        fread(&mSize.x, 4, 1, handle);
        fread(&mSize.y, 4, 1, handle);

        // allocate tiles
        if (mTiles != NULL)
            delete[] mTiles;

        mTiles = new CTile[mSize.x * mSize.y];

        // map data
        for (int y = 0; y < mSize.y; ++y)
            for (int x = 0; x < mSize.x; ++x)
            {
                CTile*  t = getTile(CVec2(x, y));
                fread(&t->mType, 1, 1, handle);

                for (int side = 0; side < eWallSideMax; ++side)
                    t->mWalls[side].load(handle);
            }

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
CParamOrnate::CParamOrnate()
{
    mType = eParamOrnate;
    mValue = 0;
}
